<?php


namespace HelloPrint\Inc\Services;

use function Automattic\WooCommerce\GoogleListingsAndAds\Vendor\GuzzleHttp\json_encode;

class HelloPrintApiService
{

    private string $base_url, $api_key, $lang, $env_mode, $callback_url;

    public function __construct($api_key = '')
    {
        $language = get_locale();
        $langArr = explode("_", $language);
        $this->lang = !empty($langArr[0]) ? $langArr[0] : 'en';
        $env_mode = get_option('helloprint_env_mode', true);
        $this->env_mode = $env_mode ? 'test' : 'prod';
        $this->base_url = HELLOPRINT_API_URL;
        $api_key ? $this->api_key = $api_key : $this->api_key = get_option('helloprint_api_key', true);
    }

    public function get($url, $params = [])
    {
        $response = wp_remote_get(
            $this->base_url . $url,
            [
                'timeout' => 45,
                'headers' => [
                    'x-api-key' => $this->api_key,
                ],
                'body' => $params,
            ]
        );
        $this->updateLangFromResponse($response);
        return $response;
    }

    public function post($url, $params, $contentType = 'application/json')
    {
        $headers = [
            'x-api-key' => $this->api_key
        ];
        if (!empty($contentType)) {
            $headers['Content-Type'] = $contentType;
        }
        $response =  wp_remote_post($this->base_url . $url, [
            'timeout' => 45,
            'headers' => $headers,
            'body' => ($contentType == 'application/json') ? \json_encode($params) : $params,
            'method' => 'POST',
            'data_format' => 'body'
        ]);
        $this->updateLangFromResponse($response);
        return $response;
    }

    public function getResponseToJson($response)
    {
        $body = wp_remote_retrieve_body($response);
        if (is_wp_error($response)) {
            return [];
        }
        if (wp_remote_retrieve_response_code($response) == '403') {
            return [];
        }
        return json_decode($body, true);
    }

    public function getCategoriesForSelectOptions()
    {
        $response = $this->get('categories');
        $data = $this->getResponseToJson($response);
        if (is_array($data) && count($data) > 0) {
            foreach ($data['data'] as $category) {
                $categories[$category] = wp_kses(_translate_helloprint(ucfirst(str_replace(['-', '_'], ' ', $category)), 'helloprint'), false);
            }
            return $categories;
        }
        return [];
    }

    public function getCategoryProductForSelectOptions($categoryId)
    {
        $response = $this->get("categories/$categoryId");
        $data = $this->getResponseToJson($response);
        if ($data) {
            if (count($data) > 0) {
                foreach ($data['data']['products'] as $category) {
                    $categories[esc_attr($category['productKey'])] = wp_kses(_translate_helloprint($category['name'], 'helloprint'), false);
                }
                return $categories;
            }
        }
        return [];
    }

    public function getProductDetailForSelectOptions($productId)
    {
        $response = $this->get("products/$productId");
        $data = $this->getResponseToJson($response);
        if ($data) {
            if (count($data) > 0) {
                $attributes = [];
                $options = [];
                $i = 0;
                $available = false;
                $destinationCountries = [];
                if (isset($data['data'])) {
                    foreach ($data['data']['attributes'] as $attributeKey => $attribute) {
                        array_push($attributes, [
                            'id' => esc_attr($attributeKey),
                            'name' => isset($attribute['name'][$this->lang]) ? esc_attr($attribute['name'][$this->lang]) : esc_attr($attribute['name']['en'])
                        ]);
                        $options = [];
                        foreach ($attribute['values'] as $value) {
                            foreach ($value as $key => $innerValue) {
                                $key = esc_attr($key);
                                $options[$key]['name'] = isset($innerValue['name'][$this->lang]) ? esc_attr($innerValue['name'][$this->lang]) : esc_attr($innerValue['name']['en']);

                                // assign sub text
                                $options[$key]['subText'] = NULL;
                                if (isset($innerValue['subText'][$this->lang]) || isset($innerValue['subText']['en'])) {
                                    $options[$key]['subText'] = isset($innerValue['subText'][$this->lang]) ? esc_attr($innerValue['subText'][$this->lang]) : esc_attr($innerValue['subText']['en']);
                                }
                                // assign image
                                $options[$key]['image'] = NULL;
                                if (isset($innerValue['image'][$this->lang]['url']) || isset($innerValue['image']['en']['url'])) {
                                    $options[$key]['image'] = isset($innerValue['image'][$this->lang]['url']) ? esc_attr($innerValue['image'][$this->lang]['url']) : esc_attr($innerValue['image']['en']['url']);
                                }
                            }
                        }
                        $attributes[$i]['options'] = $options;
                        $i++;
                    }

                    $options = (isset($data['data']['options'])) ? $data['data']['options'] : [];
                    $available = isset($data['data']['available']) ? $data['data']['available'] : false;

                    if (isset($data['data']['destinationCountries']) && count($data['data']['destinationCountries'])) {
                        foreach ($data['data']['destinationCountries'] as $kk => $dest) {
                            array_push($destinationCountries, [
                                'id' => esc_attr($kk),
                                'name' => isset($dest['name'][$this->lang]) ? esc_attr($dest['name'][$this->lang]) : esc_attr($dest['name']['en']),
                                'flag_icon' => esc_url($dest['flagIcon']),
                                'code' => esc_attr($dest['isoCode'])
                            ]);
                        }
                    }
                }
                return ['attributes' => $attributes, 'options' => $options, 'available' => $available, 'destination_countries' => $destinationCountries];
            }
        }
        return [];
    }


    public function getProductVariantsFilter($externalProductId, $attributes, $productMargin, $deliveryType = '', $selectedAttribute = '', $productQuantity = '', $productId = null, $apprealQuantity = -1, $destination_country = '')
    {
        $product = wc_get_product($productId);
        if (!empty($product) && wc_tax_enabled() && get_option("woocommerce_tax_display_shop") == 'incl' && $product->get_tax_status() == 'taxable') {
            $tax = new \WC_Tax();
            $taxes = $tax->get_rates($product->get_tax_class());
            $rates = array_shift($taxes);
        }
        $taxRate = isset($rates['rate']) ? $rates['rate'] : 0;
        $selectedPosition = array_search($selectedAttribute, array_keys($attributes));
        $query = '';
        $originalAttributeOptions = $attributes['helloprint_options'];
        $originalAttributes =  $this->getProductDetailForSelectOptions($externalProductId);
        $oriAttrOptions = isset($originalAttributes['options']) ? $originalAttributes['options'] : [];
        $newAttributes = [];
        $filterAttribute = [];
        $markup = ((100 - (float)$productMargin) / 100);
        $j = 0;
        $attributesKeys = [];
        $attributeOptions = [];
        $attributeOptionsK = [];
        $attributeFilters = [];
        if (isset($attributes['helloprint_options'])) {
            $customOptions = $attributes['helloprint_options'];
            unset($attributes['helloprint_options']);
        }
        $emptyAttr = true;
        foreach ($attributes as $vall) {
            if (!empty($vall) && ($vall != '0' || $vall != 0)) {
                $emptyAttr = false;
            }
        }
        $oriAttrs = isset($originalAttributes['attributes']) ? $originalAttributes['attributes'] : [];
        $attributesKeys = array_keys($attributes);
        $ifSelected = empty($oriAttrs) || !$emptyAttr;
        foreach ($attributes as $key => $attribute) {
            $j = array_search($key, $attributesKeys);
            if ($selectedAttribute) {
                if ($selectedPosition >= $j && $attribute != '0') {
                    $query = $query . 'attributes["' . $key . '"]="' . $attribute . '"&';
                    $filterAttribute[$key] = $attribute;
                } else {
                    array_push($newAttributes, $key);
                }
            } else {
                if ($attribute != '0') {
                    $filterAttribute[$key] = $attribute;
                    $query = $query . 'attributes["' . $key . '"]="' . $attribute . '"&';
                } else {
                    array_push($newAttributes, $key);
                }
            }
            if ($j > 0 && isset($attributeFilters[$j - 1])) {
                $attributeFilters[$j] = $attributeFilters[$j - 1];
            }
            if ($attribute != '0') {
                $attributeFilters[$j][$key] = $attribute;
            }
            //$attributesKeys[$j] = $key;
            //$j++;
        }

        if ($apprealQuantity >= 0) {
            foreach ($filterAttribute as $ky => $val) {
                if (str_starts_with($ky, 'appreal_')) {
                    unset($filterAttribute[$ky]);
                }
            }
        }
        $response = $this->get("products/$externalProductId/variants", ["where" => $query]);
        $data = $this->getResponseToJson($response);
        if ($data) {
            if (isset($data['data']) && is_countable($data['data']) && count($data['data']) > 0) {
                $variants = [];
                $quoteVariants = [];
                $originalSku = '';
                $newVariants = [];
                $i = 0;
                if (isset($data['data']['variantKey'])) {
                    $quoteVariants[$i]['variantKey'] = esc_attr($data['data']['variantKey']);
                }
                foreach ($data['data'] as $variant) {
                    if (!empty($variant['attributes'])) {
                        if (count(array_diff($filterAttribute, $variant['attributes'])) == 0) {
                            if ($i == 0 && isset($variant['variantKey'])) {
                                $quoteVariants[$i]['variantKey'] = esc_attr($variant['variantKey']);
                            }
                            if ($i == 0 && isset($variant['sku'])) {
                                $originalSku = esc_attr($variant['sku']);
                            }
                            $i++;

                            foreach ($variant['attributes'] as $key => $attribute) {
                                if (in_array($key, $newAttributes)) {
                                    $variants[$key][$attribute] = esc_attr($attribute);
                                }
                            }
                        }

                        foreach ($variant['attributes'] as $key => $attribute) {
                            $arrKey = array_search($key, $attributesKeys, true);
                            $newAttrArray = isset($attributeFilters[$arrKey - 1]) ? $attributeFilters[$arrKey - 1] : null;
                            if ($arrKey > 0 && !empty($newAttrArray)) {
                                $flag = true;
                                foreach ($newAttrArray as $nKey => $att) {
                                    if (strtoupper(trim($variant['attributes'][$nKey])) !== strtoupper(trim($att))) {
                                        $flag = false;
                                    }
                                }
                                if ($flag == true) {
                                    $attributeOptionsK[$key][$attribute] = esc_attr($attribute);
                                }
                            } else {
                                $attributeOptionsK[$key][$attribute] = esc_attr($attribute);
                            }

                            if (!isset($attributeOptionsK[$key])) {
                                $attributeOptionsK[$key] = [];
                            }
                        }
                    }
                }
                if (isset($originalAttributes['attributes'])) {
                    $originalAttributes = $originalAttributes['attributes'];
                }

                foreach ($originalAttributes as $orgKey => $originalAttribute) {
                    foreach ($variants as $key => $variant) {
                        if ($originalAttribute['id'] == $key) {
                            foreach ($originalAttribute['options'] as $optionKey => $orgOption) {
                                $search = array_search($optionKey, $variant);
                                if ($search) {
                                    if (isset($orgOption['name'])) {
                                        $orgOption['name'] = wp_kses(_translate_helloprint($orgOption['name'], 'helloprint'), false);
                                    }
                                    $newVariants[$key][$search] = $orgOption;
                                }
                            }
                        }
                    }

                    foreach ($attributeOptionsK as $key => $opt) {
                        if ($originalAttribute['id'] == $key) {
                            foreach ($originalAttribute['options'] as $optionKey => $orgOption) {
                                $search = array_search($optionKey, $opt);
                                if ($search) {
                                    $attributeOptions[$key][$search] = esc_attr($search);
                                }
                            }
                        }
                    }
                }
                $emptyService = empty($deliveryType);
                $deliveryType = empty($deliveryType) ? 'saver' : $deliveryType;
                if (!empty($customOptions)) {
                    $options = [];
                    foreach ($customOptions as $key => $opt) {
                        $options[] = ['code' => $key, 'value' => $opt];
                    }
                    $quoteVariants[0]['options'] = $options;
                }

                if ($apprealQuantity > 0) {
                    $quoteVariants[0]['quantity'] = $apprealQuantity;
                }

                $min_max_qty_message = '';
                if (!empty($originalSku) && !empty($oriAttrOptions)) {
                    $responsevar = $this->get("products/$externalProductId/variants", ["sku" => $originalSku, 'includeAvailableQtys' => true]);
                    $datavar = $this->getResponseToJson($responsevar);
                    if (isset($datavar['data']['availableQtys']) && count($datavar['data']['availableQtys']) > 0) {
                        $qts = $datavar['data']['availableQtys'];
                        $count = count($datavar['data']['availableQtys']);
                        $max = max(array_column($qts, 'quantity'));
                        $min = min(array_column($qts, 'quantity'));
                        $min_max_qty_message = wp_kses(_translate_helloprint("Min:", 'helloprint'), false) . " " . $min . "/" . wp_kses(_translate_helloprint("Max:", 'helloprint'), false)  . " " . $max;
                    }
                }
                $widthExists = array_key_exists('width', $originalAttributeOptions);
                $getQuantity = $widthExists ? (!empty($originalAttributeOptions['width']) && !empty($originalAttributeOptions['height'])) : true;

                if ($ifSelected && count($newVariants) <= 0 && $apprealQuantity <> 0 && $getQuantity) {
                    $res = $this->returnsQuotesServiceLevel($emptyService, $deliveryType, $quoteVariants, $destination_country);
                    if (isset($res)) {
                        if (isset($res['data'])) {
                            if (count($res['data']['items']) > 0) {
                                $serviceLevel = [];
                                $quotes = [];
                                $quantities = [];
                                foreach ($res['data']['items'] as $variants) {
                                    foreach ($variants as $variant) {
                                        foreach ($variant as $item) {
                                            $quantities[$item['quantity']]['quantity'] = $item['quantity'];
                                            $originalCentAmountExclTax = (float)$item['prices']['centAmountExclTax'];
                                            $originalCentAmountInclTax = $originalCentAmountExclTax * (1 + ($taxRate / 100));
                                            if ($markup > 0) {
                                                $quantities[$item['quantity']]['centAmountExclTax'] = wc_price(round($originalCentAmountExclTax / $markup) / 100);
                                                $quantities[$item['quantity']]['centAmountInclTax'] = wc_price(round($originalCentAmountInclTax / $markup) / 100);

                                                // set original value without margin  to test 
                                                $quantities[$item['quantity']]['originalcentAmtExclTax'] = wc_price($originalCentAmountExclTax / 100);
                                                $quantities[$item['quantity']]['originalcentAmountInclTax'] = wc_price($originalCentAmountInclTax / 100);
                                            } else {
                                                $quantities[$item['quantity']]['centAmountExclTax'] = wc_price($originalCentAmountExclTax / 100);
                                                $quantities[$item['quantity']]['centAmountInclTax'] = wc_price($originalCentAmountInclTax / 100);
                                            }
                                        }
                                    }
                                }
                                $attributePosition = array_search($selectedAttribute, array_column($originalAttributes, 'id'));
                                $attributesCount = count($attributes) - 1;
                                if (count($quantities) > 0) {
                                    asort($quantities);
                                    if ($productQuantity == '') {
                                        if ($selectedAttribute == '' || $attributePosition == $attributesCount) {
                                            $productQuantity = array_values($quantities)[0]['quantity'];
                                        }
                                    }
                                }
                                if (isset($quoteVariants[0]['serviceLevel'])) {
                                    unset($quoteVariants[0]['serviceLevel']);
                                }

                                $quoteReqParams = ['items' => $quoteVariants];
                                if (!empty($destination_country)) {
                                    $quoteReqParams['destinationCountryCode'] = $destination_country;
                                }
                                $serviceLevelRes =  $this->post('quotes', $quoteReqParams);

                                $serviceLevelRes =  $this->getResponseToJson($serviceLevelRes);
                                if (isset($serviceLevelRes['data'])) {
                                    foreach ($serviceLevelRes['data']['items'] as $key => $variants) {
                                        foreach ($variants as $variantkey => $quoteDetails) {
                                            foreach ($quoteDetails as $itemKey => $item) {
                                                if ($item['serviceLevel'] && !empty($item['serviceLevel']) && $item['serviceLevel'] != null) {
                                                    $serviceLabelObj = [];
                                                    if (isset($item['times']) && !empty($item['times'])) {
                                                        $serviceLabelObj = [
                                                            'label' => wp_kses(_translate_helloprint(ucwords($item['serviceLevel']), "helloprint"), true) . ' - ' . $item['times']['minDeliveryDays'] . ' ' . wp_kses(_translate_helloprint("Day(s)", 'helloprint'), true),
                                                            'value' => $item['serviceLevel'],
                                                            'days' => $item['times']['minDeliveryDays']
                                                        ];
                                                    } else {
                                                        $serviceLabelObj = [
                                                            'label' => wp_kses(_translate_helloprint(ucwords($item['serviceLevel']), "helloprint"), true),
                                                            'value' => $item['serviceLevel'],
                                                            'days' => 0
                                                        ];
                                                    }

                                                    if (!in_array($serviceLabelObj, $serviceLevel) && (array_search($item['serviceLevel'], array_column($serviceLevel, 'value')) === FALSE)) {
                                                        array_push($serviceLevel, $serviceLabelObj);
                                                    }
                                                    if ($item['serviceLevel'] == $deliveryType) {
                                                        $quotes[] = $item;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    array_multisort(array_map(function ($element) {
                                        return $element['days'];
                                    }, $serviceLevel), SORT_NUMERIC, $serviceLevel);
                                }

                                foreach ($quotes as $key => $quote) {
                                    $centAmountExclTax = (float)$quote['prices']['centAmountExclTax'];
                                    $centAmountInclTax = $centAmountExclTax * (1 + ($taxRate / 100));
                                    if ($markup > 0) {
                                        $quotes[$key]['prices']['centAmountExclTax'] = round((float)$quote['prices']['centAmountExclTax'] / $markup);
                                        $quotes[$key]['prices']['centAmountInclTax'] = round($centAmountInclTax / $markup);

                                        // set original value without margin  to test 
                                        $quotes[$key]['prices']['origcentAmountExclTax'] = round((float)$quote['prices']['centAmountExclTax']);
                                        $quotes[$key]['prices']['origcentAmountInclTax'] = round($centAmountInclTax);
                                    }

                                    if (empty($quotes[$key]['serviceLevel']) || $quotes[$key]['serviceLevel'] == null) {
                                        unset($quotes[$key]);
                                        array_splice($quotes, 0, 0);
                                    }
                                }

                                return [
                                    'quotes' => $quotes,
                                    'quantities' => $quantities,
                                    'currency' => $res['data']['currency'],
                                    'serviceLevel' => $serviceLevel,
                                    'variants' => $newVariants,
                                    'attributeOptions' => $attributeOptions,
                                    'tax_rate' => $taxRate,
                                    'min_max_qty_message' => !empty($min_max_qty_message) ? $min_max_qty_message : ''
                                ];
                            }
                        }
                    }
                }
                return [
                    'quotes' => [],
                    'quantities' => [],
                    'currency' => '',
                    'serviceLevel' => [],
                    'variants' => $newVariants,
                    'attributeOptions' => $attributeOptions,
                    'tax_rate' => $taxRate,
                    'min_max_qty_message' => !empty($min_max_qty_message) ? $min_max_qty_message : ''
                ];
            }
        }
        return [
            'quotes' => [],
            'quantities' => [],
            'currency' => '',
            'serviceLevel' => [],
            'variants' => [],
            'attributeOptions' => $attributeOptions,
            'tax_rate' => $taxRate,
            'min_max_qty_message' => !empty($min_max_qty_message) ? $min_max_qty_message : ''
        ];
    }

    private function returnsQuotesServiceLevel($emptyService, &$deliveryType, $quoteVariants, $destination_country = '')
    {

        $res =  $this->sendRequestToQuote($quoteVariants, $deliveryType, $destination_country);
        if (empty($res['data']) && ($emptyService || $deliveryType == 'saver')) {
            if (!isset($res['data']) || empty($res['data'])) {
                $deliveryType = 'standard';
                $res =  $this->sendRequestToQuote($quoteVariants, $deliveryType, $destination_country);
            }

            if (!isset($res['data']) || empty($res['data'])) {
                $deliveryType = 'express';
                $res =  $this->sendRequestToQuote($quoteVariants, $deliveryType, $destination_country);
            }
        }
        return $res;
    }

    private function sendRequestToQuote($quoteVariants, $deliveryType, $destination_country = '')
    {
        $quoteVariants[0]['serviceLevel'] = $deliveryType;
        $quoteParams = [
            'items' => $quoteVariants
        ];
        if (!empty($destination_country)) {
            $quoteParams['destinationCountryCode'] = $destination_country;
        }
        $quoteRes =  $this->post('quotes', $quoteParams);
        return $this->getResponseToJson($quoteRes);
    }

    public function createOrder($order)
    {
        $logger = wc_get_logger();
        $order_json = json_decode($order, true);
        $orderItems = [];

        foreach ($order->get_items() as $key => $item) {
            $itemId = (int)esc_attr($item->get_id());
            $product = wc_get_product($item['product_id']);
            $item = json_decode($item['helloprint_product_setup'], true);
            $fileItems = '';
            $zipname = 'helloprint-wp-' . $order->get_id() . '-' . $itemId . '.zip';
            $variantKey = ($item['helloprint_variant_key']) ?? '';
            $quantity = ($item['quantity']) ?? '';
            $serviceLevel = ($item['delivery_option']) ?? '';
            $sku = !empty(esc_attr($item['sku'])) ? esc_attr($item['sku']) : esc_attr($product->get_sku());          
            global $wpdb;
            $preset_tableName = $wpdb->prefix . 'helloprint_order_presets';
            $line_item_tableName = $wpdb->prefix . 'helloprint_order_line_item_presets';
            $line_item_public_file_tableName = $wpdb->prefix . 'helloprint_order_line_public_files';
            $line_item_file_tableName = $wpdb->prefix . 'helloprint_order_line_preset_files';
            $lineItemPreset = $wpdb->get_results("SELECT $line_item_tableName.*, $preset_tableName.helloprint_variant_key, $preset_tableName.file_url  from $line_item_tableName INNER JOIN $preset_tableName ON $line_item_tableName.preset_id = $preset_tableName.id where $line_item_tableName.line_item_id = '$itemId' and ($preset_tableName.helloprint_item_sku = '$sku' OR $preset_tableName.helloprint_item_sku = '')");

            $lineItemPreset = ($lineItemPreset[0]) ?? [];

            if (empty($lineItemPreset) && !empty($sku)) {
                $lineItemPreset = $wpdb->get_results("SELECT $preset_tableName.*  from $preset_tableName where helloprint_item_sku = '$sku'");
                $lineItemPreset = ($lineItemPreset[0]) ?? [];
                if (!empty($lineItemPreset)) {
                    $lineItemPreset->service_level = $lineItemPreset->default_service_level;
                    $lineItemPreset->quantity = $lineItemPreset->default_quantity;
                    $presetDefaultFile = $lineItemPreset->file_url;
                } else {
                    $logger->info("Place order: No preset file for order item" .  wp_json_encode($product));
                }
            }
            if (!empty($lineItemPreset)) {
                if (!empty($lineItemPreset->helloprint_variant_key))
                    $variantKey = $lineItemPreset->helloprint_variant_key;
                if (!empty($lineItemPreset->service_level))
                    $serviceLevel = $lineItemPreset->service_level;
                if (!empty($lineItemPreset->quantity))
                    $quantity = $lineItemPreset->quantity;
            }

            $homePath = ABSPATH;
            $baseFolder = $homePath . 'wp-content/uploads/print-tmp/';
            if (!file_exists($baseFolder)) {
                mkdir($baseFolder, 0755, true);
                $logger->info('Place order: folder created for zip file');
            }
            $zipPath = $baseFolder . $zipname;
            if (file_exists($zipPath)) {
                unlink($zipPath);
                $logger->info('Place order: Duplicate zip file deleted');
            }

            $zip = new \ZipArchive;
            $zip->open($zipPath, \ZIPARCHIVE::CREATE);
            $logger->info('Place order: Zip file opened');
            $fileCount = 0;
            
            $presetFiles = $wpdb->get_results("SELECT * from $line_item_file_tableName where line_item_id = '$itemId'");
            if (!empty($presetFiles)) {
                foreach ($presetFiles as $key => $file) {
                    if ($file->file_url != '') {
                        $fileUrl = $homePath . ltrim($file->file_url, "/");
                        $zip->addFile($fileUrl, basename($fileUrl));
                        $logger->info('Place order: ' . $file->file_url . " file added to zip");
                        $fileCount++;
                    }
                }
            } else if(!empty($presetDefaultFile)) {
                $fileUrl = $homePath . ltrim($presetDefaultFile, "/");
                $zip->addFile($fileUrl, basename($fileUrl));
                $fileCount++;
            }

            $zip->close();
            if ($fileCount > 0) {
                $fileItems = \get_site_url() . '/wp-content/uploads/print-tmp/' . $zipname;
                $logger->info('Place order: Zip file created');
            }

            $publicFiles = $wpdb->get_results("SELECT * from $line_item_public_file_tableName where line_item_id = '$itemId'");
            if (!empty($publicFiles[0]->public_file_url)) {
                if (!empty($fileItems)) {
                    $fileItems .= ", ";
                }
                $fileItems .= $publicFiles[0]->public_file_url;
            }
            
            $pp_item_string = \wc_get_order_item_meta($itemId, '_w2p_set_option', true);
            if (!empty($pp_item_string)) {
                $json = \urldecode($pp_item_string);
                $array = \json_decode($json, true);
                if (!empty($array['distiller']) && !empty($array['projectId'])) {
                    $pp_url = $array['distiller'] . "/?id=" . $array['projectId'];
                    if (!empty($fileItems)) {
                        $fileItems .= ", ";
                    }
                    $fileItems .= $pp_url;
                }
            }
            if (!empty($variantKey) && !empty($quantity)) {
                $orderItem = [
                    'itemReferenceId' => 'wp-' . $order_json['id'] . '-' . $itemId,
                    'variantKey' => $variantKey,
                    'quantity' => $quantity,
                    'serviceLevel' => $serviceLevel,
                    'fileUrl' => $fileItems
                ];

                if (!empty($item['custom_options'])) {
                    $options = [];
                    if (!is_array($item['custom_options'])) {
                        $item['custom_options'] = json_decode($item['custom_options'], true);
                    }
                    foreach ($item['custom_options'] as $key => $opt) {
                        $options[] = ['code' => strtolower($key), 'value' => $opt];
                    }
                    $orderItem['options'] = $options;
                }

                if (!empty($item['options']['appreal_size'])) {
                    $apprealSizes = $item['options']['appreal_size'];
                    $finalSizesString = "";
                    foreach ($apprealSizes as $key => $opt) {
                        $finalSizesString .= strtolower($key) . "[";
                        $interArr = [];
                        foreach ($opt as $oKey => $oVal) {
                            if ($oVal > 0) {
                                $interArr[] = strtolower($oKey) . ":" . $oVal;
                            }
                        }
                        $finalSizesString .= implode("; ", $interArr);
                        $finalSizesString .= "]";
                    }
                    if (!isset($orderItem['options'])) {
                        $orderItem['options'] = [];
                    }
                    $appSize = ['code' => 'apparelSize', 'value' => trim($finalSizesString)];
                    $orderItem['options'][] = $appSize;
                }
                $logger->info('Place order: Order item detail: ' . wp_json_encode($orderItem));
                $orderItems[] = $orderItem;
            }
        }
        $shipOrbilling = $order_json['shipping']['first_name'] ? 'shipping' : 'billing';
        $callBackUrl = get_site_url() . '/?rest_route=/helloprint/v1/complete_order_from_helloprint_callback';
        $shippingDetails = [
            'companyName' => ($order_json[$shipOrbilling]['company']) ?? '',
            'firstName' => $order_json[$shipOrbilling]['first_name'],
            'lastName' => $order_json[$shipOrbilling]['last_name'],
            'addressLine1' => $order_json[$shipOrbilling]['address_1'],
            'addressLine2' => $order_json[$shipOrbilling]['address_2'],
            'postcode' => $order_json[$shipOrbilling]['postcode'],
            'city' => $order_json[$shipOrbilling]['city'],
            'country' => $order_json[$shipOrbilling]['country'],
            'phone' => !empty($order_json[$shipOrbilling]['phone']) ? $order_json[$shipOrbilling]['phone'] : (isset($order_json['billing']['phone']) ? $order_json['billing']['phone'] : ''),
            'email' => ($order_json['billing']['email']) ?? '',
        ];
        $logger->info('Place order: Shipping detail: ' . wp_json_encode($shippingDetails));
        if (true === WP_DEBUG) {
            error_log("order Items details :: " . \json_encode($orderItems));
            error_log("shipping details :: " . \json_encode($shippingDetails));
        }
        if (empty($orderItems) || empty($order_json[$shipOrbilling])) {
            return ['status' => 'VALIDATION_ERROR'];
        }
        $logger->info('Place order: Order placed');
        return $this->post('orders', [
            'mode' => $this->env_mode,
            'orderReferenceId' => 'wp-' . $order_json['id'],
            'shipping' => $shippingDetails,
            'orderItems' => $orderItems,
            'callbackUrls' => [
                $callBackUrl
            ]
        ]);
    }

    public function getProductDetailsForLoad($productId)
    {
        $response = $this->get("products/$productId");
        $data = $this->getResponseToJson($response);

        if ($data) {
            if (count($data) > 0 && !empty($data["data"])) {
                $returnData = [];
                $returnData['description'] = isset($data['data']['description'][$this->lang]) ? esc_attr($data['data']['description'][$this->lang]) : esc_attr($data['data']['description']['en']);
                $returnData['product_name'] = isset($data['data']['productName'][$this->lang]) ? esc_attr($data['data']['productName'][$this->lang]) : esc_attr($data['data']['productName']['en']);
                $returnData['preview_image'] = NULL;
                $returnData['gallery_images'] = NULL;
                if (!empty($data['data']['images'])) {
                    $returnData['preview_image'] = isset($data['data']['images'][0]['url']) ? esc_url($data['data']['images'][0]['url']) : NULL;
                    foreach ($data['data']['images'] as $k => $img) {
                        if (isset($img['url'])) {
                            $returnData['gallery_images'][$k] = esc_url($img['url']);
                        }
                    }
                }
                return $returnData;
            }
        }
        return [];
    }

    public function getOrderDetails($requestId)
    {
        $response = $this->get("orders/requestId=" . $requestId);
        $data = $this->getResponseToJson($response);
        return $data;
    }

    private function updateLangFromResponse($response)
    {
        if (!empty(wp_remote_retrieve_header($response, 'x-language-default'))) {
            $this->lang = wp_remote_retrieve_header($response, 'x-language-default') ? wp_remote_retrieve_header($response, 'x-language-default') : 'en';
        }
    }

    public function getAllProducts()
    {
        $response = $this->get("products");
        $data = $this->getResponseToJson($response);
        $products = [];
        if ($data) {
            if (count($data) > 0) {
                foreach ($data['data'] as $product) {
                    $productName = isset($product['productName'][$this->lang]) ? $product['productName'][$this->lang] : $product['productName']['en'];
                    $category = isset($product['categoryName'][$this->lang]) ? $product['categoryName'][$this->lang] : $product['categoryName']['en'];
                    $category = esc_attr(trim(ucwords(str_replace("-", " ", $category))));
                    if (!empty($productName)) {
                        $products[$category][esc_attr($product['productKey'])] = esc_attr($productName);
                    }
                }
                ksort($products);
                return $products;
            }
        }
        return [];
    }

    public function cancelOrder($orderId = '')
    {
        return $this->post('orders/' . $orderId . '/cancel', []);
    }

    public function getProductPdfTemplates($productId)
    {
        $response = $this->get("products/$productId/templates");
        $data = $this->getResponseToJson($response);
        if ($data) {
            if (count($data) > 0) {
                $templates = [];
                $i = 0;
                if (isset($data['data'])) {
                    foreach ($data['data'] as $key => $temp) {
                        $single_template = [];
                        $single_template['name'] = isset($temp['name'][$this->lang]) ? $temp['name'][$this->lang] : $temp['name']['en'];
                        if (!empty($temp['pdf']['url'])) {
                            $pdf_url = $temp['pdf']['url'];
                            $pdf_file_name = $temp['pdf']['fileName'];
                            if (is_array($pdf_url)) {
                                $pdf_url = isset($pdf_url[$this->lang]) ? $pdf_url[$this->lang] : $pdf_url['en'];
                            }
                            if (is_array($pdf_file_name)) {
                                $pdf_file_name = isset($pdf_file_name[$this->lang]) ? $pdf_file_name[$this->lang] : $pdf_file_name['en'];
                            }
                            $single_template['pdf']['url'] = esc_url($pdf_url);
                            $single_template['pdf']['file_name'] = esc_attr($pdf_file_name);
                        }

                        if (!empty($temp['indd']['url'])) {
                            $indd_url = $temp['indd']['url'];
                            $indd_file_name = $temp['indd']['fileName'];
                            if (is_array($indd_url)) {
                                $indd_url = isset($indd_url[$this->lang]) ? $indd_url[$this->lang] : $indd_url['en'];
                            }
                            if (is_array($indd_file_name)) {
                                $indd_file_name = isset($indd_file_name[$this->lang]) ? $indd_file_name[$this->lang] : $indd_file_name['en'];
                            }
                            $single_template['indd']['url'] = esc_url($indd_url);
                            $single_template['indd']['file_name'] = esc_attr($indd_file_name);
                        }
                        array_push($templates, $single_template);
                    }
                    return $templates;
                }
            }
        }

        return [];
    }

    public function validateApiKey($orderId = '')
    {
        return $this->get('identity?src=wordpress');
    }

    public function get_qts_and_levels($variant_key = '')
    {
        $serviceLevelRes =  $this->post('quotes', [
            'items' => [
                0 => [
                    'variantKey' => $variant_key
                ]
            ]
        ]);
        $quantities = [];
        $serviceLevels = [];
        $sku = '';
        $serviceLevelRes =  $this->getResponseToJson($serviceLevelRes);
        if (isset($serviceLevelRes['data'])) {
            foreach ($serviceLevelRes['data']['items'] as $key => $variants) {
                foreach ($variants as $variantkey => $quoteDetails) {
                    foreach ($quoteDetails as $itemKey => $item) {
                        if (!empty($item['serviceLevel'])) {
                            $serviceLevels[esc_attr($item['serviceLevel'])] =  esc_attr(ucwords($item['serviceLevel']));
                        }

                        if (!empty($item['quantity'])) {
                            $quantities[] =  esc_attr($item['quantity']);
                        }

                        if (!empty($item['sku'])) {
                            $sku =  esc_attr($item['sku']);
                        }
                    }
                }
            }
        }

        return ['quantities' => array_unique($quantities), 'service_levels' => array_unique($serviceLevels), 'sku' => $sku];
    }

    public function getAllSKUsByProductKey($externalProductId)
    {
        $response = $this->get("products/$externalProductId/variants", ["where" => 'attributes']);
        $data = $this->getResponseToJson($response);
        if ($data) {
            if (isset($data['data']) && is_countable($data['data']) && count($data['data']) > 0) {
                return $data['data'];                
            }
        }
        return [];
    }
}
